;; Generated package description from jami-bot.el  -*- no-byte-compile: t -*-
(define-package "jami-bot" "0.0.4.0.20240204.184721" "An extendable chat bot for the private messenger GNU Jami" '((emacs "27.1")) :commit "ac024faefa4b1de302b3c576188e47e2000c8376" :authors '(("Hanno Perrey" . "hanno@hoowl.se")) :maintainer '("Hanno Perrey" . "hanno@hoowl.se") :keywords '("comm" "jami" "messenger" "chat bot" "dbus") :url "https://gitlab.com/hperrey/jami-bot")
